"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const dc_configs_1 = require("dc-configs");
const dc_logging_1 = require("dc-logging");
const timers_1 = require("timers");
const events_1 = require("events");
const DAppPlayerInstance_1 = require("./DAppPlayerInstance");
const DAppDealerInstance_1 = require("./DAppDealerInstance");
const Utils = __importStar(require("dc-ethereum-utils"));
const log = new dc_logging_1.Logger("DAppInstance");
const SERVER_APPROVE_AMOUNT = 100000000;
const SERVER_APPROVE_MINAMOUNT = 10000000;
class DApp extends events_1.EventEmitter {
    constructor(params) {
        super();
        if (!params.slug) {
            log.debug(["Create DApp error", params], "error");
            throw new Error("slug option is required");
        }
        if (!params.gameContractAddress) {
            throw new Error("gameContract is not specified in  DApp params");
        }
        this._params = params;
        this._instancesMap = new Map();
        const { gameContractAddress, slug } = this._params;
        const gameId = slug;
        this._gameInfo = {
            gameId,
            slug,
            hash: Utils.checksum(slug),
            contract: gameContractAddress
        };
        this._gameContractAddress = gameContractAddress;
        this._gameContractInstance = this._params.Eth.initContract(dc_configs_1.config.default.contracts.Game.abi, gameContractAddress);
        this._gameInfoRoomAddress = `${params.platformId}_${params.blockchainNetwork}_${this._gameInfo.hash}`;
    }
    getView() {
        return { name: this._params.slug };
    }
    getInstancesView() {
        return Array.from(this._instancesMap.values()).map(instance => instance.getView());
    }
    eventNames() {
        return ["ready"];
    }
    async startClient() {
        this._gameInfoRoom = await this._params.roomProvider.getRemoteInterface(this._gameInfoRoomAddress);
        const readyServers = new Map();
        const prommise = new Promise((resolve, reject) => {
            this._gameInfoRoom.on("ready", async (readyInfo) => {
                readyServers.set(readyInfo.address, readyInfo);
                if (this._dappInstancePromise)
                    await this._dappInstancePromise;
                this._dappInstancePromise = this._chooseServer(readyServers);
                const result = await this._dappInstancePromise;
                if (result) {
                    resolve(result);
                }
            });
        });
        return prommise;
    }
    async _chooseServer(readyServers) {
        const self = this;
        if (this.dappInstance)
            return this.dappInstance;
        const theChosen = Array.from(readyServers.values())
            .filter(readyServer => {
            return readyServer.deposit;
        })
            .sort((a, b) => {
            return a.deposit - b.deposit[0];
        });
        // TODO: should be some more complicated alg
        if (theChosen) {
            const userId = this._params.Eth.getAccount().address;
            const { roomAddress } = await this._gameInfoRoom.connect({
                userId
            });
            this.dappInstance = new DAppPlayerInstance_1.DAppPlayerInstance({
                userId,
                num: 0,
                rules: this._params.rules,
                gameContractInstance: this._gameContractInstance,
                gameContractAddress: this._gameContractAddress,
                gameLogicFunction: this._params.gameLogicFunction,
                roomProvider: this._params.roomProvider,
                roomAddress,
                onFinish: this.onGameFinished,
                gameInfo: this._gameInfo,
                Eth: this._params.Eth
            });
            this.dappInstance.on("info", data => {
                self.emit("dapp::status", {
                    message: "dapp instance message",
                    data
                });
            });
            await this.dappInstance.start();
            return this.dappInstance;
        }
        self.emit("dapp::status", { message: "Server not chosen", data: {} });
        log.debug("Server not choosen");
        return null;
    }
    async startServer() {
        await this._params.roomProvider.exposeSevice(this._gameInfoRoomAddress, this, true);
        await this._params.Eth.ERC20ApproveSafe(this._gameContractAddress, SERVER_APPROVE_AMOUNT, SERVER_APPROVE_MINAMOUNT);
        return this._startSendingBeacon(3000);
    }
    async _startSendingBeacon(timeOut) {
        const { balance } = await this._params.Eth.getBetBalance(this._params.Eth.getAccount().address);
        const self = this;
        this._beaconInterval = timers_1.setInterval(() => {
            self.emit("ready", {
                // deposit: bet2dec(balance),
                deposit: balance,
                dapp: {
                    slug: self._params.slug,
                    hash: self._gameInfo.hash
                },
                address: this._params.Eth.getAccount().address.toLowerCase()
            });
        }, timeOut);
    }
    // User connect
    onGameFinished(userId) {
        this._instancesMap.delete(userId);
    }
    connect(params) {
        const roomAddress = Utils.makeSeed();
        const { userId } = params;
        const account = this._params.Eth.getAccount();
        const dappInstance = new DAppDealerInstance_1.DAppDealerInstance({
            userId,
            num: 0,
            rules: this._params.rules,
            roomAddress,
            gameLogicFunction: this._params.gameLogicFunction,
            gameContractInstance: this._gameContractInstance,
            gameContractAddress: this._gameContractAddress,
            roomProvider: this._params.roomProvider,
            onFinish: this.onGameFinished,
            gameInfo: this._gameInfo,
            Eth: this._params.Eth
        });
        dappInstance.start();
        // TODO remove circular dependency
        this._instancesMap.set(userId, dappInstance);
        log.debug(`User ${userId} connected to  ${this._params.slug}`);
        return { roomAddress };
    }
}
exports.DApp = DApp;
//# sourceMappingURL=DApp.js.map